import logging
import os
import sys
from logging.config import fileConfig, dictConfig

import gunicorn.glogging
from gunicorn.glogging import CONFIG_DEFAULTS


class GunicornLogger(gunicorn.glogging.Logger):
    def __init__(self, cfg):
        super().__init__(cfg)

    def setup(self, cfg):
        self.loglevel = self.LOG_LEVELS.get(cfg.loglevel.lower(), logging.INFO)
        self.error_log.setLevel(self.loglevel)
        self.access_log.setLevel(logging.INFO)

        # set gunicorn.error handler
        if self.cfg.capture_output and cfg.errorlog != "-":
            for stream in sys.stdout, sys.stderr:
                stream.flush()

            self.logfile = open(cfg.errorlog, "a+")
            os.dup2(self.logfile.fileno(), sys.stdout.fileno())
            os.dup2(self.logfile.fileno(), sys.stderr.fileno())

        # [DCSC] Force gunicorn to produce an access log; without pushing this to stdout from this instance
        self.cfg.set("accesslog", "Y")

        # [DCSC] Removed error log handler from this instance
        # self._set_handler(
        #     self.error_log,
        #     cfg.errorlog,
        #     logging.Formatter(self.error_fmt, self.datefmt),
        # )

        # [DCSC] Removed stdout accesslog handler
        # set gunicorn.access handler
        # if cfg.accesslog is not None:
        #     self._set_handler(
        #         self.access_log,
        #         cfg.accesslog,
        #         fmt=logging.Formatter(self.access_fmt),
        #         stream=sys.stdout,
        #     )

        # [DCSC] Removed syslog handler
        # set syslog handler
        # if cfg.syslog:
        #     self._set_syslog_handler(self.error_log, cfg, self.syslog_fmt, "error")
        #     if not cfg.disable_redirect_access_to_syslog:
        #         self._set_syslog_handler(
        #             self.access_log, cfg, self.syslog_fmt, "access"
        #         )

        if cfg.logconfig_dict:
            config = CONFIG_DEFAULTS.copy()
            config.update(cfg.logconfig_dict)
            try:
                dictConfig(config)
            except (AttributeError, ImportError, ValueError, TypeError) as exc:
                raise RuntimeError(str(exc))
        elif cfg.logconfig:
            if os.path.exists(cfg.logconfig):
                defaults = CONFIG_DEFAULTS.copy()
                defaults["__file__"] = cfg.logconfig
                defaults["here"] = os.path.dirname(cfg.logconfig)
                fileConfig(
                    cfg.logconfig, defaults=defaults, disable_existing_loggers=False
                )
            else:
                msg = "Error: log config '%s' not found"
                raise RuntimeError(msg % cfg.logconfig)



import json
from collections import namedtuple
import io
from json import JSONDecodeError
from typing import Any

import requests
from requests import Response
from requests.adapters import HTTPAdapter, Retry


class ApiBaseClass(object):
    """
    The GenericApi class serves as a base class for all API's
    """

    def __init__(
        self,
        baseurl: str,
        api_path: str = None,
        proxies: dict = None,
        user_agent: str = "ApiBaseClass",
        **kwargs,
    ):
        """
        The Generic api caller handles all communication towards an api resource.
        """

        if "verify" not in kwargs:
            self.verify = False
        else:
            self.verify = kwargs["verify"]
            kwargs.pop("verify")

        self.baseurl = baseurl
        self.api_path = api_path
        self.proxies = proxies
        self.user_agent = user_agent
        self.kwargs = kwargs

        self.methods = namedtuple("methods", "GET POST DELETE PUT PATCH")(
            "GET", "POST", "DELETE", "PUT", "PATCH"
        )

        self.myheaders = self.__default_headers

    def __repr__(self) -> str:
        """return a string representation of the obj GenericApi"""
        return f"<< GenericApi:{self.baseurl} >>"

    def _build_url(self, resource: str) -> str:
        """
        Internal method to build a url to use when executing commands
        """
        if self.api_path is None:
            return f"{self.baseurl}/{resource}"
        else:
            return f"{self.baseurl}/{self.api_path}/{resource}"

    def _connect(
        self,
        method: str,
        resource: str,
        session: requests.Session,
        data: dict | list | str | bytes | io.TextIOBase = None,
        timeout: int = 60,
        return_response_object: bool = False,
        stream: bool = False,
    ) -> Response | str | Any:
        """
        Send a request

        Send a request to api host based on the specified data. Specify the content type as JSON and
        convert the data to JSON format.
        """

        requests.packages.urllib3.disable_warnings()

        request_api_resource = {
            "headers": self.myheaders,
            "verify": self.verify,
            "timeout": timeout,
            "proxies": self.proxies,
        }

        if data is not None:
            if method in [self.methods.DELETE, self.methods.GET]:
                if not isinstance(data, (list, dict, bytes)):
                    raise TypeError(
                        f"'data' type for {method} must be dict, bytes or a list of tuples."
                    )
                request_api_resource["params"] = data
            else:
                if isinstance(data, dict):
                    try:
                        data = json.dumps(data)
                    except:
                        raise TypeError(
                            "Dict provided to 'data' is not json serializable."
                        )
                elif not isinstance(data, (str, list, bytes, io.TextIOBase)):
                    raise TypeError(
                        f"'data' type for {method} must be str, dict, bytes, file-like or a list of tuples."
                    )

                request_api_resource["data"] = data

        request_api_resource.update(self.kwargs)

        try:
            if method == self.methods.POST:
                r = session.post(
                    self._build_url(resource), stream=stream, **request_api_resource
                )
            elif method == self.methods.PUT:
                r = session.put(
                    self._build_url(resource), stream=stream, **request_api_resource
                )
            elif method == self.methods.PATCH:
                r = session.patch(
                    self._build_url(resource), stream=stream, **request_api_resource
                )
            elif method == self.methods.DELETE:
                r = session.delete(
                    self._build_url(resource), stream=stream, **request_api_resource
                )
            else:
                r = session.get(
                    self._build_url(resource), stream=stream, **request_api_resource
                )

            try:
                if isinstance(r, Response):
                    if return_response_object or stream:
                        return r
                    if r.status_code >= 400:
                        if isinstance(r.text, str):
                            raise requests.exceptions.ConnectionError(r.text)
                        else:
                            the_response = json.loads(r.text)
                            raise requests.exceptions.ConnectionError(the_response)
                    else:
                        the_response = json.loads(r.text)
                else:
                    the_response = r
            except JSONDecodeError:
                if "content-type" in r.headers:
                    if r.headers["content-type"] == "text/plain":
                        the_response = r.text
                    else:
                        the_response = r
                else:
                    the_response = r

            return the_response
        except requests.exceptions.ConnectionError as err:
            raise requests.exceptions.ConnectionError(err)
        except Exception as err:
            raise Exception(err)

    def get_session(
        self,
        retries: int = 1,
        backoff_factor: float = 0.3,
        status_forcelist: tuple = (429, 500, 502, 503, 504),
        session=None,
    ) -> requests.Session:
        """
        Method for returning a session object per every requesting thread
        """
        session = session or requests.Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount("http://", adapter)
        session.mount("https://", adapter)

        return session

    def call(
        self,
        method: str = None,
        resource: str = None,
        data: dict = None,
        timeout: int = 60,
        return_response_object: bool = False,
        stream: bool = False,
    ) -> Response | str | Any:
        """
        Method for requesting free format api resources

        """
        try:
            with self.get_session() as session:
                result = self._connect(
                    method=method,
                    resource=resource,
                    session=session,
                    data=data,
                    timeout=timeout,
                    return_response_object=return_response_object,
                    stream=stream,
                )
                return result
        except requests.ConnectionError:
            raise
        except Exception:
            raise

    @property
    def __default_headers(self):
        """
        Property to return the default headers

        :return: Default headers
        :rtype: dict
        """
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": f"{self.user_agent}",
        }

    @property
    def headers(self) -> dict:
        """
        Property to return the current headers
        """

        return self.myheaders

    def set_header_field(self, field: str, value: str) -> dict:
        """
        Method to add a header and set it's value
        """

        self.myheaders[field] = value

        return self.myheaders

    def del_header_field(self, field: str) -> dict:
        """
        Method to delete a header field
        """

        self.myheaders.pop(field)

        return self.myheaders

    def reset_headers(self):
        """
        Method to reset the headers to the default values
        """

        self.myheaders = self.__default_headers

    def clear_headers(self):
        """
        Method to clear the headers and the myheaders to None
        """
        self.myheaders = None



import logging
from typing import List, Tuple

from flask import Flask

from nldcsc.flask_midddleware.base import BaseHTTPMiddleware
from nldcsc.loggers.app_logger import AppLogger

logging.setLoggerClass(AppLogger)


class MiddlewareManager(object):
    def __init__(self, app: Flask = None):
        self.app = app
        self._middleware_stack = {}
        self.logger = logging.getLogger(self.__class__.__name__)

        if self.app is not None:
            self.init_app(app=app)

    @property
    def middleware_stack(self) -> List[Tuple[int, dict]]:
        return sorted(self._middleware_stack.items(), reverse=True)

    def add_middleware(self, middleware_class: type, rank: int, **options):
        """

        Args:
            middleware_class: middleware class
            rank: middleware prority; lower numbered ranks are preferred over higher numbered ranks
            **options: options for the middleware class initialization

        Returns:

        """
        if BaseHTTPMiddleware.__subclasscheck__(middleware_class):
            self._middleware_stack[rank] = {"cls": middleware_class, "options": options}
        else:
            raise Exception(
                "Error Middleware Class : not inherited from BaseHTTPMiddleware class"
            )

    def init_app(self, app: Flask):
        for mw_entry in self.middleware_stack:
            rank, mw = mw_entry
            # init the middleware
            mw["cls"](**mw["options"])

        app.middleware_manager = self

    def __repr__(self):
        return "<< MiddlewareManager >>"



import os
from json import JSONDecodeError

import pytest

from nldcsc.generic.utils import (
    getenv_dict,
    getenv_bool,
    getenv_list,
    str2bool,
    generate_random_password,
)


class TestGeneric:
    def test_get_env_dict(self):
        os.environ["DICT_ENV"] = '{"key1": "value1", "key2": "value2"}'
        os.environ["DICT_ENV_ERROR"] = "{'key1': 'value1', 'key2': 'value2'}"

        env_dict = getenv_dict("DICT_ENV", {"key": "default"})

        assert isinstance(env_dict, dict)
        assert env_dict["key1"] == "value1"

        env_dict = getenv_dict("DICT_ENV_NON_EXISTENT", {"key": "default"})

        assert isinstance(env_dict, dict)
        assert env_dict["key"] == "default"

        env_dict = getenv_dict("DICT_ENV_NON_EXISTENT")

        assert isinstance(env_dict, dict)
        assert len(env_dict) == 0

        with pytest.raises(JSONDecodeError):
            getenv_dict("DICT_ENV_ERROR", {"key": "default"})

    def test_get_env_list(self):
        os.environ["LIST_ENV"] = '["value1", "value2"]'
        os.environ["LIST_ENV_ERROR"] = "['value1', 'value2']"

        env_list = getenv_list("LIST_ENV", ["default"])

        assert isinstance(env_list, list)
        assert env_list[0] == "value1"

        env_list = getenv_list("LIST_ENV_NON_EXISTENT", ["default"])

        assert isinstance(env_list, list)
        assert env_list[0] == "default"

        env_list = getenv_list("LIST_ENV_NON_EXISTENT")

        assert isinstance(env_list, list)
        assert len(env_list) == 0

        with pytest.raises(JSONDecodeError):
            getenv_list("LIST_ENV_ERROR", ["default"])

    def test_get_env_bool(self):
        os.environ["BOOL_ENV_TRUE"] = "True"
        os.environ["BOOL_ENV_FALSE"] = "False"
        os.environ["BOOL_ENV_ERROR"] = "Blerf"
        os.environ["BOOL_ENV_DICT"] = '{"key1": "value1", "key2": "value2"}'

        env_bool = getenv_bool("BOOL_ENV_TRUE", "False")
        env_bool_false = getenv_bool("BOOL_ENV_FALSE", "False")

        assert isinstance(env_bool, bool)
        assert env_bool is True
        assert env_bool_false is False

        env_bool = getenv_bool("BOOL_ENV_NON_EXISTENT", "False")

        assert isinstance(env_bool, bool)
        assert env_bool is False

        with pytest.raises(ValueError):
            getenv_bool("BOOL_ENV_ERROR", "False")

        with pytest.raises(ValueError):
            getenv_bool("BOOL_ENV_DICT", "False")

    def test_str_to_bool(self):
        choices = {"yes", "true", "1"}

        assert str2bool(choices.pop()) is True
        assert str2bool("false") is False

    def test_random_password_generator(self):
        assert len(generate_random_password()) == 32



import logging
from collections import defaultdict

from elasticsearch import BadRequestError
from flask import request

from nldcsc.datatables.server_side_dt import ServerSideDataTable
from nldcsc.flask_plugins.flask_eswrap import FlaskEsWrap
from nldcsc.generic.utils import str2bool
from nldcsc.loggers.app_logger import AppLogger

logging.setLoggerClass(AppLogger)


class ESServerSideDataTable(ServerSideDataTable):
    def __init__(self, request: request, backend: FlaskEsWrap, index: str, **kwargs):
        self.index = index
        self.logger = logging.getLogger(__name__)

        super().__init__(request, backend, **kwargs)

    def get_results_and_meta(self):
        self.total = getattr(self.backend.es, self.index).count()

        self.filtered = self.data_filter()

        self.fetch_results()

        if len(self.filtered) != 0:
            try:
                self.total_filtered = getattr(self.backend.es, self.index).count(
                    filter=self.filtered, regexp=True, bool_query=True
                )
            except BadRequestError as err:
                self.logger.warning(
                    f"Received BadRequestError from elasticsearch -> {err}"
                )
                self.logger.exception(err)
                self.total_filtered = 0
        else:
            self.total_filtered = self.total

    def fetch_results(self):
        try:
            if len(self.filtered) != 0:
                data_objects = list(
                    getattr(self.backend.es, self.index)
                    .find(filter=self.filtered, regexp=True, bool_query=True)
                    .skip(self.data_length.start)
                    .limit(self.data_length.length)
                )
            else:
                data_objects = list(
                    getattr(self.backend.es, self.index)
                    .find()
                    .skip(self.data_length.start)
                    .limit(self.data_length.length)
                )
        except BadRequestError as err:
            self.logger.warning(f"Received BadRequestError from elasticsearch -> {err}")
            self.logger.exception(err)
            data_objects = []

        self.results = data_objects

    def post_fetch_processing(self):
        pass

    def data_filter(self):
        docfilter = defaultdict(list)

        search_val = self.request_values["search[value]"]

        if search_val != "" and search_val != "[":
            column_search_list = [
                self.columns[i]["data"]
                for i in self.columns
                if str2bool(self.columns[i]["searchable"])
            ]

            for each in column_search_list:
                docfilter["should"].append({"regexp": {each: search_val}})

        return dict(docfilter)
